#include "MPU6886.h"
#include <Kalman.h>
#include <WiFi.h>
#include <WebServer.h>
#include <Preferences.h>
#include <M5GFX.h>

M5GFX lcd;

WebServer server(80);

const char ssid[] = "S-3 DC";  // SSID
const char pass[] = "password";   // password

const IPAddress ip(192, 168, 22, 33);      // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク

#define Lch1 5
#define Lch2 4
#define Rch1 3
#define Rch2 2
#define Cch1 1
#define Cch2 0

#define Lin1 5
#define Lin2 6
#define Rin1 2
#define Rin2 1
#define Cin1 7
#define Cin2 8

MPU6886 IMU;

unsigned long oldTime = 0, loopTime, nowTime;
float dt;

float omegaZ = 0.0, varW = 0.0;
int omZon = 0;

float Prot = 0.1;
float Kp = 50;
float Kd = 70;
float Kw = 2.0;
float IDRS = 2.5;
int delayTime = 0;

int pwmDutyC, pwmDutyL, pwmDutyR;
int GetUP = 0;
int GetUpCnt = 0;
float MtC, MtL, MtR, MtX, MtY, MpreX = 0.0, MpreY = 0.0;
float AjX = 0.0, AjY = 0.0;

float accX = 0, accY = 0, accZ = 0;
float gyroX = 0, gyroY = 0, gyroZ = 0;

float theta_X = 0.0, theta_Y = 0.0;
float theta_Xdot = 0.0, theta_Ydot = 0.0;

Kalman kalmanX, kalmanY;
float kalAngleX, kalAngleDotX, kalAngleY, kalAngleDotY;

float offsetX = 0.0, offsetY = 0.0;

Preferences preferences;


//加速度センサから傾きデータ取得 [deg]
void get_theta_acc() {
  IMU.getAccelData(&accX, &accY, &accZ);
  //傾斜角導出 単位はdeg
  theta_X  = atan2(accY, accZ) * 57.29578f + offsetX;
  theta_Y  = atan2(-accX, accZ) * 57.29578f + offsetY;
}

//Y軸 角速度取得
void get_gyro_data() {
  IMU.getGyroData(&gyroX, &gyroY, &gyroZ);
  theta_Xdot = gyroX;
  theta_Ydot = gyroY;
}



//ブラウザ表示
void handleRoot() {
  String temp = "<!DOCTYPE html> \n<html lang=\"ja\">";
  temp += "<head>";
  temp += "<meta charset=\"utf-8\">";
  temp += "<title>SHISEIGYO-3 DC</title>";
  temp += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
  temp += "<style>";
  temp += ".container{";
  temp += "  max-width: 500px;";
  temp += "  margin: auto;";
  temp += "  text-align: center;";
  temp += "  font-size: 1.2rem;";
  temp += "}";
  temp += "span,.pm{";
  temp += "  display: inline-block;";
  temp += "  border: 1px solid #ccc;";
  temp += "  width: 50px;";
  temp += "  height: 30px;";
  temp += "  vertical-align: middle;";
  temp += "  margin-bottom: 20px;";
  temp += "}";
  temp += "span{";
  temp += "  width: 120px;";
  temp += "}";
  temp += "button{";
  temp += "  width: 100px;";
  temp += "  height: 40px;";
  temp += "  font-weight: bold;";
  temp += "  margin-bottom: 20px;";
  temp += "}";
  temp += "</style>";
  temp += "</head>";

  temp += "<body>";
  temp += "<div class=\"container\">";
  temp += "<h3>SHISEIGYO-3 DC</h3>";

  //omegaZ
  temp += "回転<br>";
  temp += "<a class=\"pm\" href=\"/omegaZm\">-</a>";
  temp += "<span>" + String(omegaZ) + "</span>";
  temp += "<a class=\"pm\" href=\"/omegaZp\">+</a><br>";

  //Prot
  temp += "回転パラメータ<br>";
  temp += "<a class=\"pm\" href=\"/ProtM\">-</a>";
  temp += "<span>" + String(Prot) + "</span>";
  temp += "<a class=\"pm\" href=\"/ProtP\">+</a><br>";


  //offsetX
  temp += "offsetX<br>";
  temp += "<a class=\"pm\" href=\"/offsetXm\">-</a>";
  temp += "<span>" + String(offsetX) + "</span>";
  temp += "<a class=\"pm\" href=\"/offsetXp\">+</a><br>";

  //offsetY
  temp += "offsetY<br>";
  temp += "<a class=\"pm\" href=\"/offsetYm\">-</a>";
  temp += "<span>" + String(offsetY) + "</span>";
  temp += "<a class=\"pm\" href=\"/offsetYp\">+</a><br>";

  

  //Kp
  temp += "Kp<br>";
  temp += "<a class=\"pm\" href=\"/KpM\">-</a>";
  temp += "<span>" + String(Kp) + "</span>";
  temp += "<a class=\"pm\" href=\"/KpP\">+</a><br>";

  //Kd
  temp += "Kd<br>";
  temp += "<a class=\"pm\" href=\"/KdM\">-</a>";
  temp += "<span>" + String(Kd) + "</span>";
  temp += "<a class=\"pm\" href=\"/KdP\">+</a><br>";

  //Kw
  temp += "Kw<br>";
  temp += "<a class=\"pm\" href=\"/KwM\">-</a>";
  temp += "<span>" + String(Kw) + "</span>";
  temp += "<a class=\"pm\" href=\"/KwP\">+</a><br>";

  //IDRS
  temp += "IDRS<br>";
  temp += "<a class=\"pm\" href=\"/IDRSm\">-</a>";
  temp += "<span>" + String(IDRS) + "</span>";
  temp += "<a class=\"pm\" href=\"/IDRSp\">+</a><br>";

  //delayTime
  temp += "delayTime<br>";
  temp += "<a class=\"pm\" href=\"/delayTimeM\">-</a>";
  temp += "<span>" + String(delayTime) + "</span>";
  temp += "<a class=\"pm\" href=\"/delayTimeP\">+</a><br>";

  temp += "</div>";
  temp += "</body>";
  server.send(200, "text/HTML", temp);
}


void omegaZm() {
  if (omegaZ >= -50) {
    omegaZ -= 5;
  }
  handleRoot();
}

void omegaZp() {
  if (omegaZ <= 50) {
    omegaZ += 5;
  }
  handleRoot();
}

void ProtM() {
  if (Prot >= 0) {
    Prot -= 0.1;
    preferences.putFloat("Prot", Prot);
  }
  handleRoot();
}

void ProtP() {
  if (Prot <= 100) {
    Prot += 0.1;
    preferences.putFloat("Prot", Prot);
  }
  handleRoot();
}


void offsetXm() {
  if (offsetX >= -10) {
    offsetX -= 0.1;
    preferences.putFloat("offsetX", offsetX);
  }
  handleRoot();
}

void offsetXp() {
  if (offsetX <= 10) {
    offsetX += 0.1;
    preferences.putFloat("offsetX", offsetX);
  }
  handleRoot();
}


void offsetYm() {
  if (offsetY >= -10) {
    offsetY -= 0.1;
    preferences.putFloat("offsetY", offsetY);
  }
  handleRoot();
}

void offsetYp() {
  if (offsetY <= 10) {
    offsetY += 0.1;
    preferences.putFloat("offsetY", offsetY);
  }
  handleRoot();
}



void KpM() {
  if (Kp >= 0.1) {
    Kp -= 1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KpP() {
  if (Kp <= 300) {
    Kp += 1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KdM() {
  if (Kd >= 0.1) {
    Kd -= 1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KdP() {
  if (Kd <= 300) {
    Kd += 1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KwM() {
  if (Kw >= 0) {
    Kw -= 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void KwP() {
  if (Kw <= 300) {
    Kw += 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void IDRSm() {
  if (IDRS >= 0.1) {
    IDRS -= 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}

void IDRSp() {
  if (IDRS <= 10) {
    IDRS += 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}


void delayTimeM() {
  if (delayTime >= 0) {
    delayTime -= 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}

void delayTimeP() {
  if (delayTime <= 30) {
    delayTime += 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}




//Core0
void display(void *pvParameters) {
  lcd.begin();
  lcd.setRotation(2);
  lcd.setTextColor(0xFFFFFFU, 0x000000U); //１つ目の引数が文字色、２つ目の引数が背景色


  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // 追記：このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定

  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  server.on("/", handleRoot);

  server.on("/omegaZm", omegaZm);
  server.on("/omegaZp", omegaZp);
  server.on("/ProtM", ProtM);
  server.on("/ProtP", ProtP);
  
  server.on("/offsetXm", offsetXm);
  server.on("/offsetXp", offsetXp);
  server.on("/offsetYm", offsetYm);
  server.on("/offsetYp", offsetYp);
  
  server.on("/KpP", KpP);
  server.on("/KpM", KpM);
  server.on("/KdP", KdP);
  server.on("/KdM", KdM);
  server.on("/KwP", KwP);
  server.on("/KwM", KwM);
  server.on("/IDRSp", IDRSp);
  server.on("/IDRSm", IDRSm);
  server.on("/delayTimeP", delayTimeP);
  server.on("/delayTimeM", delayTimeM);

  server.begin();

  for (;;) {
    server.handleClient();

    lcd.setTextFont(4);
    lcd.setCursor(0, 1);
    lcd.printf("%+05.1f", kalAngleX);
    //lcd.printf("%+07.5f", varW);
    lcd.setCursor(64, 1);
    lcd.printf("%+05.1f", kalAngleY);

    int xx = map(-kalAngleX, 20, -20, 25, 127);
    xx = constrain(xx, 25, 127);
    uint32_t colorX;

    if (fabs(kalAngleX) <= 1.0) {
      colorX = 0x00FF00U;
    } else if (fabs(kalAngleX) <= 15) {
      colorX = 0x0000FFU;
    } else {
      colorX = 0xFF0000U;
    }
    lcd.fillRect(0, xx - 2, 128, 5, colorX);


    int yy = map(-kalAngleY, 20, -20, 0, 127);
    yy = constrain(yy, 0, 127);
    uint32_t colorY;

    if (fabs(kalAngleY) <= 1.0) {
      colorY = 0x00FF00U;
    } else if (fabs(kalAngleY) <= 15) {
      colorY = 0x0000FFU;
    } else {
      colorY = 0xFF0000U;
    }
    lcd.fillRect(yy - 2, 25, 5, 128, colorY);
    delay(33);
    lcd.clear();

    /*
      USBSerial.print(rotEst);
      USBSerial.print(", ");
      USBSerial.println(rotSpeed);
    */

    disableCore0WDT();
  }
}



void setup() {
  Serial.begin(115200);

  IMU.Init();

  //フルスケールレンジ
  IMU.SetAccelFsr(IMU.AFS_2G);
  IMU.SetGyroFsr(IMU.GFS_250DPS);

  get_theta_acc();
  kalmanX.setAngle(theta_X);
  kalmanY.setAngle(theta_Y);

  ledcSetup(Lch1, 20000, 10);
  ledcAttachPin(Lin1, Lch1);
  ledcSetup(Lch2, 20000, 10);
  ledcAttachPin(Lin2, Lch2);

  ledcSetup(Rch1, 20000, 10);
  ledcAttachPin(Rin1, Rch1);
  ledcSetup(Rch2, 20000, 10);
  ledcAttachPin(Rin2, Rch2);

  ledcSetup(Cch1, 20000, 10);
  ledcAttachPin(Cin1, Cch1);
  ledcSetup(Cch2, 20000, 10);
  ledcAttachPin(Cin2, Cch2);

  ledcWrite(Lch1, 1023);
  ledcWrite(Lch2, 1023);
  ledcWrite(Rch1, 1023);
  ledcWrite(Rch2, 1023);
  ledcWrite(Cch1, 1023);
  ledcWrite(Cch2, 1023);

  //パラメータ初期値取得
  preferences.begin("parameter", false);

  Prot = preferences.getFloat("Prot", Prot);
  
  offsetX = preferences.getFloat("offsetX", offsetX);
  offsetY = preferences.getFloat("offsetY", offsetY);
  
  Kp = preferences.getFloat("Kp", Kp);
  Kd = preferences.getFloat("Kd", Kd);
  Kw = preferences.getFloat("Kw", Kw);
  IDRS = preferences.getFloat("IDRS", IDRS);
  delayTime = preferences.getInt("delayTime", delayTime);


  //回転位置検出 タスク
  xTaskCreatePinnedToCore(
    display
    ,  "display"   // A name just for humans
    ,  4096  // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  0);
}


void loop() {
  nowTime = micros();
  loopTime = nowTime - oldTime;
  oldTime = nowTime;

  dt = (float)loopTime / 1000000.0; //sec


  get_theta_acc();
  get_gyro_data();

  //カルマンフィルタ 姿勢 傾き
  kalAngleX = kalmanX.getAngle(theta_X, theta_Xdot, dt);
  kalAngleY = kalmanY.getAngle(theta_Y, theta_Ydot, dt);

  //カルマンフィルタ 姿勢 角速度
  kalAngleDotX = kalmanX.getRate();
  kalAngleDotY = kalmanY.getRate();


  if (fabs(kalAngleX) < 1 && fabs(kalAngleY) < 1 && GetUP == 0) {
    GetUP = 80;
  }


  if (GetUP == 80) {
    //ブレーキ
    if (fabs(kalAngleX) > 15.0 || fabs(kalAngleY) > 15.0) {
      ledcWrite(Lch1, 1023);
      ledcWrite(Lch2, 1023);
      ledcWrite(Rch1, 1023);
      ledcWrite(Rch2, 1023);
      ledcWrite(Cch1, 1023);
      ledcWrite(Cch2, 1023);
      AjX = 0.0;
      AjY = 0.0;
      varW = 0.0;
      GetUP = 0;
    }

    //モータ回転
    if (GetUP == 80) {
      AjY +=  IDRS * MpreY / 100.0;
      MtY = Kp * (kalAngleY + AjY) / 90.0 + Kd * kalAngleDotY / 500.0 + Kw * MpreY / 10.0;

      AjX +=  IDRS * MpreX / 100.0;
      MtX = Kp * (kalAngleX + AjX) / 90.0 + Kd * kalAngleDotX / 500.0 + Kw * MpreX / 10.0;
    }


    //回転速度追従ON
    if(omegaZ != 0.0){
      omZon = 1;
      varW = Prot / 100.0 * (omegaZ - gyroZ);
    }

    //回転速度追従OFF
    if(fabs(varW) < 0.000005 && omZon == 1 && fabs(MtR) < 0.001){
      omZon = 0;
      varW = 0.0;
    }


    MtC = constrain(MtY + varW , -1.0, 1.0);
    MpreY = MtC;
    pwmDutyC = 1023 * (1.0 - fabs(MtC));
    if (pwmDutyC < 300) pwmDutyC = 300;
    
    
    MtL = constrain(float(0.5 * MtY - 0.866 * MtX), -1.0, 1.0);
    pwmDutyL = 1023 * (1.0 - fabs(MtL));
    if (pwmDutyL < 300) pwmDutyL = 300;

    
    MtR = constrain(float(0.5 * MtY + 0.866 * MtX), -1.0, 1.0);
    MpreX = (MtR - MtL) * 0.5;
    pwmDutyR = 1023 * (1.0 - fabs(MtR));
    if (pwmDutyR < 300) pwmDutyR = 300;

    
    
    //回転方向
    if (MtC > 0.0) {
      ledcWrite(Cch1, pwmDutyC);
      ledcWrite(Cch2, 1023);
    } else {
      ledcWrite(Cch1, 1023);
      ledcWrite(Cch2, pwmDutyC);
    }
    
    if (MtL < 0.0) {
      ledcWrite(Lch1, pwmDutyL);
      ledcWrite(Lch2, 1023);
    } else {
      ledcWrite(Lch1, 1023);
      ledcWrite(Lch2, pwmDutyL);
    }

    if (MtR < 0.0) {
      ledcWrite(Rch1, pwmDutyR);
      ledcWrite(Rch2, 1023);
    } else {
      ledcWrite(Rch1, 1023);
      ledcWrite(Rch2, pwmDutyR);
    }
  } else if (GetUP == 0) {
    ledcWrite(Lch1, 1023);
    ledcWrite(Lch2, 1023);
    ledcWrite(Rch1, 1023);
    ledcWrite(Rch2, 1023);
    ledcWrite(Cch1, 1023);
    ledcWrite(Cch2, 1023);
  }

  delay(delayTime);
}
